# -*- coding: utf-8 -*-
# Copyright 2020 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The Bgp_global parser templates file. This contains
a list of parser definitions and associated functions that
facilitates both facts gathering and native command generation for
the given network resource.
"""

import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


def _tmplt_router_bgp_cmd(config_data):
    command = "router bgp {as_number}".format(**config_data)
    return command


def _tmplt_bgp_vrf(config_data):
    command = "vrf {vrf}".format(**config_data)
    return command


def _tmplt_bgp_aggregate_address(config_data):
    command = "aggregate-address {address}".format(**config_data)
    if config_data.get("as_set"):
        command += " as-set"
    if config_data.get("summary_only"):
        command += " summary-only"
    if config_data.get("attribute_map"):
        command += " attribute-map {attribute_map}".format(**config_data)
    if config_data.get("match_map"):
        command += " match-map {match_map}".format(**config_data)
    if config_data.get("advertise_only"):
        command += " advertise-only"
    return command


def _tmplt_bgp_params(config_data):
    command = "bgp"
    if config_data["bgp_params"].get("additional_paths"):
        command += " additional-paths " + config_data["bgp_params"]["additional_paths"]
        if config_data["bgp_params"]["additional_paths"] == "send":
            command += " any"
    elif config_data["bgp_params"].get("advertise_inactive"):
        command += " advertise-inactive"
    elif config_data["bgp_params"].get("allowas_in"):
        command += " allowas-in"
        if config_data["bgp_params"]["allowas_in"].get("count"):
            command += " {count}".format(**config_data["bgp_params"]["allowas_in"])
    elif config_data["bgp_params"].get("always_compare_med"):
        command += " always-compare-med"
    elif config_data["bgp_params"].get("asn"):
        command += " asn notaion {asn}".format(**config_data["bgp_params"])
    elif config_data["bgp_params"].get("auto_local_addr"):
        command += " auto-local-addr"
    elif config_data["bgp_params"].get("bestpath"):
        if config_data["bgp_params"]["bestpath"].get("as_path"):
            command += " bestpath as-path {as_path}".format(**config_data["bgp_params"]["bestpath"])
        elif config_data["bgp_params"]["bestpath"].get("ecmp_fast"):
            command += " bestpath ecmp-fast"
        elif config_data["bgp_params"]["bestpath"].get("med"):
            command += " bestpath med"
            if config_data["bgp_params"]["bestpath"]["med"].get("confed"):
                command += " confed"
            else:
                command += " missing-as-worst"
        elif config_data["bgp_params"]["bestpath"].get("skip"):
            command += " bestpath skip next-hop igp-cost"
        elif config_data["bgp_params"]["bestpath"].get("tie_break"):
            tie = re.sub(
                r"_",
                r"-",
                config_data["bgp_params"]["bestpath"]["tie_break"],
            )
            command += " tie-break " + tie
    elif config_data["bgp_params"].get("client_to_client"):
        command += " client-to-client reflection"
    elif config_data["bgp_params"].get("cluster_id"):
        command += " cluster-id {cluster_id}".format(**config_data["bgp_params"])
    elif config_data["bgp_params"].get("confederation"):
        command += " confederation"
        if config_data["bgp_params"]["confederation"].get("identifier"):
            command += " identifier " + config_data["bgp_params"]["confederation"]["identifier"]
        else:
            command += " peers {peers}".format(**config_data["bgp_params"]["confederation"])
    elif config_data["bgp_params"].get("control_plane_filter"):
        command += " control-plane-filter default-allow"
    elif config_data["bgp_params"].get("convergence"):
        command += " convergence"
        if config_data["bgp_params"]["convergence"].get("slow_peer"):
            command += " slow-peer"
        command += " time {time}".format(**config_data["bgp_params"]["convergence"])
    elif config_data["bgp_params"].get("default"):
        command += " default {default}".format(**config_data["bgp_params"])
    elif config_data["bgp_params"].get("enforce_first_as"):
        command += " enforce-first-as"
    elif config_data["bgp_params"].get("host_routes"):
        command += " host-routes fib direct-install"
    elif config_data["bgp_params"].get("labeled_unicast"):
        command += " labeled-unicast rib {labeled_unicast}".format(**config_data["bgp_params"])
    elif config_data["bgp_params"].get("listen"):
        # from eos 4.23 , 'bgp listen limit ' is replaced by 'dynamic peer max'.
        command = "dynamic peer max "
        if config_data["bgp_params"]["listen"].get("limit"):
            command += "{limit}".format(**config_data["bgp_params"]["listen"])
        else:
            command += " range {address} peer group".format(
                **config_data["bgp_params"]["listen"]["range"],
            )
            if config_data["bgp_params"]["listen"]["range"]["peer_group"].get(
                "peer_filter",
            ):
                command += " {name} peer-filter {peer_filter}".format(
                    **config_data["bgp_params"]["listen"]["range"]["peer_group"],
                )
            else:
                command += " {name} remote-as {remote_as}".format(
                    **config_data["bgp_params"]["listen"]["range"]["peer_group"],
                )
    elif config_data["bgp_params"].get("log_neighbor_changes"):
        command += " log-neighbor-changes"
    elif config_data["bgp_params"].get("missing_policy"):
        command += " missing-policy direction {direction} action {action}".format(
            **config_data["bgp_params"]["missing_policy"],
        )
    elif config_data["bgp_params"].get("monitoring"):
        command += " monitoring"
    elif config_data["bgp_params"].get("next_hop_unchanged"):
        command += " next-hop-unchanged"
    elif config_data["bgp_params"].get("redistribute_internal"):
        command += " redistribute-internal"
    elif config_data["bgp_params"].get("route"):
        command += " route install-map {route}".format(**config_data["bgp_params"])
    elif config_data["bgp_params"].get("route_reflector"):
        command += " route-reflector preserve-attributes"
        if config_data["bgp_params"]["route_reflector"].get("preserve"):
            command += " always"
    elif config_data["bgp_params"].get("transport"):
        command += " transport listen-port {transport}".format(**config_data["bgp_params"])
    return command


def _tmplt_bgp_redistribute(config_data):
    command = "redistribute {protocol}".format(**config_data)
    if config_data.get("isis_level"):
        command += " {isis_level}".format(**config_data)
    if config_data.get("ospf_route"):
        if config_data["ospf_route"] == "nssa_external_2":
            route = "nssa-external 2"
        elif config_data["ospf_route"] == "nssa_external_1":
            route = "nssa-external 1"
        else:
            route = config_data["ospf_route"]
        command += " match " + route
    if config_data.get("route_map"):
        command += " route-map {route_map}".format(**config_data)
    return command


def _tmplt_bgp_default_metric(config_data):
    command = "default-metric {default_metric}".format(**config_data)
    return command


def _tmplt_bgp_distance(config_data):
    command = "distance bgp"
    if config_data["distance"].get("external"):
        command += " {external}".format(**config_data["distance"])
    if config_data["distance"].get("internal"):
        command += " {internal}".format(**config_data["distance"])
    if config_data["distance"].get("local"):
        command += " {local}".format(**config_data["distance"])
    return command


def _tmplt_bgp_graceful_restart(config_data):
    command = "graceful-restart"
    if config_data.get("restart_time"):
        command += " restart-time {restart_time}".format(**config_data)
    if config_data.get("stalepath_time"):
        command += " stalepath-time {stalepath_time}".format(**config_data)
    return command


def _tmplt_bgp_graceful_restart_helper(config_data):
    command = "graceful-restart-helper"
    return command


def _tmplt_bgp_access_group(config_data):
    if config_data.get("afi") == "ipv4":
        afi = "ip"
    else:
        afi = "ipv6"
    command = afi + " access-group {acl_name}".format(**config_data)
    if config_data.get("direction"):
        command += " {direction}".format(**config_data)
    return command


def _tmplt_bgp_maximum_paths(config_data):
    command = "maximum-paths {max_equal_cost_paths}".format(**config_data["maximum_paths"])
    if config_data["maximum_paths"].get("max_installed_ecmp_paths"):
        command += " ecmp {max_installed_ecmp_paths}".format(**config_data["maximum_paths"])
    return command


def _tmplt_bgp_monitoring(config_data):
    cmd = "monitoring"
    command = ""
    if config_data.get("timestamp"):
        command = cmd + " timestamp {timestamp}".format(**config_data)
    if config_data.get("port"):
        command = cmd + " port {port}".format(**config_data)
    if config_data.get("received"):
        command = cmd + " received routes {received}".format(**config_data)
    if config_data.get("station"):
        command = cmd + " station {station}".format(**config_data)
    return command


def _tmplt_bgp_neighbor(config_data):
    command = "neighbor {neighbor_address}".format(**config_data["neighbor"])
    if config_data["neighbor"].get("additional_paths"):
        command += " additional-paths {additional_paths}".format(**config_data["neighbor"])
        if config_data["neighbor"]["additional_paths"] == "send":
            command += "any"
    elif config_data["neighbor"].get("peer_group"):
        command += " peer group"
        if config_data["neighbor"]["peer_group"] != config_data["neighbor"]["peer_group"]:
            command += config_data["neighbor"]["peer_group"]
    elif config_data["neighbor"].get("allowas_in"):
        command += " allowas-in"
        if config_data["neighbor"]["allowas_in"].get("count"):
            command += " {count}".format(**config_data["neighbor"]["allowas_in"])
    elif config_data["neighbor"].get("auto_local_addr"):
        command += " auto-local-addr"
    elif config_data["neighbor"].get("bfd"):
        command += " bfd"
        if config_data["neighbor"]["bfd"] == "c_bit":
            command += " c-bit"
    elif config_data["neighbor"].get("default_originate"):
        command += " default-originate"
        if config_data["neighbor"]["default_originate"].get("route_map"):
            command += " route-map {route_map}".format(
                **config_data["neighbor"]["default_originate"],
            )
        if config_data["neighbor"]["default_originate"].get("always"):
            command += " always"
    elif config_data["neighbor"].get("description"):
        command += " description {description}".format(**config_data["neighbor"])
    elif config_data["neighbor"].get("dont_capability_negotiate"):
        command += " dont-capability-negotiate"
    elif config_data["neighbor"].get("ebgp_multihop"):
        command += " ebgp-multiphop"
        if config_data["neighbor"]["ebgp_multihop"].get("ttl"):
            command += " {ttl}".format(**config_data["neighbor"]["ebgp_multihop"])
    elif config_data["neighbor"].get("encryption_password"):
        command += " password {type} {password}".format(
            **config_data["neighbor"]["encryption_password"],
        )
    elif config_data["neighbor"].get("enforce_first_as"):
        command += " enforce-first-as"
    elif config_data["neighbor"].get("export_localpref"):
        command += " export-localpref {export_localpref}".format(**config_data["neighbor"])
    elif config_data["neighbor"].get("fall_over"):
        command += " fall-over bfd"
    elif config_data["neighbor"].get("graceful_restart"):
        command += " graceful-restart"
    elif config_data["neighbor"].get("graceful_restart_helper"):
        command += " graceful-restart-helper"
    elif config_data["neighbor"].get("idle_restart_timer"):
        command += " idle-restart-timer {idle_restart_timer}".format(**config_data["neighbor"])
    elif config_data["neighbor"].get("import_localpref"):
        command += " import-localpref {import_localpref}".format(**config_data["neighbor"])
    elif config_data["neighbor"].get("link_bandwidth"):
        command += " link-bandwidth"
        if config_data["neighbor"]["link_bandwidth"].get("auto"):
            command += " auto"
        if config_data["neighbor"]["link_bandwidth"].get("default"):
            command += " default {default}".format(**config_data["neighbor"]["link_bandwidth"])
        if config_data["neighbor"]["link_bandwidth"].get("update_delay"):
            command += " update-delay {update_delay}".format(
                **config_data["neighbor"]["link_bandwidth"],
            )
    elif config_data["neighbor"].get("local_as"):
        command += " local-as {as_number} no-prepend replace-as".format(
            **config_data["neighbor"]["local_as"],
        )
        if config_data["neighbor"]["local_as"].get("fallback"):
            command += " fallback"
    elif config_data["neighbor"].get("local_v6_addr"):
        command += " local-v6-addr {local_v6_addr}".format(**config_data["neighbor"])
    elif config_data["neighbor"].get("maximum_accepted_routes"):
        command += " maximum-accepted-routes {count}".format(
            **config_data["neighbor"]["maximum_accepted_routes"],
        )
        if config_data["neighbor"]["maximum_accepted_routes"].get(
            "warning_limit",
        ):
            command += " warning-limit {warning_limit}".format(
                **config_data["neighbor"]["maximum_accepted_routes"],
            )
    elif config_data["neighbor"].get("maximum_received_routes"):
        command += " maximum-routes {count}".format(
            **config_data["neighbor"]["maximum_received_routes"],
        )
        if config_data["neighbor"]["maximum_received_routes"].get(
            "warning_limit",
        ):
            if config_data["neighbor"]["maximum_received_routes"]["warning_limit"].get(
                "limit_count",
            ):
                command += " warning-limit {limit_count}".format(
                    **config_data["neighbor"]["maximum_received_routes"]["warning_limit"],
                )
            if config_data["neighbor"]["maximum_received_routes"]["warning_limit"].get(
                "limit_percent",
            ):
                command += (
                    " warning-limit "
                    + str(
                        config_data["neighbor"]["maximum_received_routes"]["warning_limit"][
                            "limit_percent"
                        ],
                    )
                    + " percent"
                )
        if config_data["neighbor"]["maximum_received_routes"].get(
            "warning_only",
        ):
            command += " warning-only"
    elif config_data["neighbor"].get("metric_out"):
        command += " metric-out {metric_out}".format(**config_data["neighbor"])
    elif config_data["neighbor"].get("monitoring"):
        command += " monitoring"
    elif config_data["neighbor"].get("next_hop_self"):
        command += " next-hop-self"
    elif config_data["neighbor"].get("next_hop_unchanged"):
        command += " next-hop-unchanged"
    elif config_data["neighbor"].get("next_hop_v6_address"):
        command += " next-hop-v6-addr {next_hop_v6_address} in".format(**config_data["neighbor"])
    elif config_data["neighbor"].get("out_delay"):
        command += " out-delay {out_delay}".format(**config_data["neighbor"])
    elif config_data["neighbor"].get("remote_as"):
        command += " remote-as {remote_as}".format(**config_data["neighbor"])
    elif config_data["neighbor"].get("remove_private_as"):
        command += " remove-private-as"
        if config_data["neighbor"]["remove_private_as"].get("all"):
            command += " all"
        if config_data["neighbor"]["remove_private_as"].get("replace_as"):
            command += " replace-as"
    elif config_data["neighbor"].get("peer_as"):
        command += " peer-as {peer_as}".format(**config_data["neighbor"])
    elif config_data["neighbor"].get("prefix_list"):
        command += " prefix-list {name} {direction}".format(
            **config_data["neighbor"]["prefix_list"],
        )
    elif config_data["neighbor"].get("route_map"):
        command += " route-map {name} {direction}".format(**config_data["neighbor"]["route_map"])
    elif config_data["neighbor"].get("route_reflector_client"):
        command += " route-reflector-client"
    elif config_data["neighbor"].get("route_to_peer"):
        command += " route-to-peer"
    elif config_data["neighbor"].get("send_community"):
        command += " send-community"
        if config_data["neighbor"]["send_community"].get(
            "community_attribute",
        ):
            command += " " + config_data["neighbor"]["send_community"]["community_attribute"]
        if config_data["neighbor"]["send_community"].get("sub_attribute"):
            command += " " + config_data["neighbor"]["send_community"]["sub_attribute"]
        if config_data["neighbor"]["send_community"].get(
            "link_bandwidth_attribute",
        ):
            command += " " + config_data["neighbor"]["send_community"]["link_bandwidth_attribute"]
        if config_data["neighbor"]["send_community"].get("speed"):
            command += " " + config_data["neighbor"]["send_community"]["speed"]
        if config_data["neighbor"]["send_community"].get("divide"):
            command += " " + config_data["neighbor"]["send_community"]["divide"]
    elif config_data["neighbor"].get("shutdown"):
        command += " shutdown"
    elif config_data["neighbor"].get("soft_reconfiguration"):
        command += " soft-reconfiguration inbound"
        if config_data["neighbor"]["soft_reconfiguration"] == "all":
            command += " all"
    elif config_data["neighbor"].get("transport"):
        command += " transport"
        if config_data["neighbor"]["transport"].get("connection_mode"):
            command += " connection-mode passive"
        else:
            command += " remote-port {remote_port}".format(**config_data["neighbor"]["transport"])
    elif config_data["neighbor"].get("timers"):
        command += " timers {keepalive} {holdtime}".format(**config_data["neighbor"]["timers"])
    elif config_data["neighbor"].get("ttl"):
        command += " ttl maximum-hops {ttl}".format(**config_data["neighbor"])
    elif config_data["neighbor"].get("update_source"):
        command += " update-source {update_source}".format(**config_data["neighbor"])
    elif config_data["neighbor"].get("weight"):
        command += " weight {weight}".format(**config_data["neighbor"])
    return command


def _tmplt_bgp_network(config_data):
    command = "network {address}".format(**config_data)
    if config_data.get("route_map"):
        command += " route-map {route_map}".format(**config_data)
    return command


def _tmplt_bgp_route_target(config_data):
    command = "route-target {action}".format(**config_data["route_target"])
    if config_data["route_target"].get("type"):
        command += " {type}".format(**config_data["route_target"])
    if config_data["route_target"].get("route_map"):
        command += " {route_map}".format(**config_data["route_target"])
    if config_data["route_target"].get("imported_route"):
        command += " imported-route"
    if config_data["route_target"].get("target"):
        command += " {target}".format(**config_data["route_target"])

    return command


def _tmplt_bgp_router_id(config_data):
    command = "router-id {router_id}".format(**config_data)
    return command


def _tmplt_bgp_shutdown(config_data):
    return "shutdown"


def _tmplt_bgp_timers(config_data):
    command = "timers bgp {keepalive} {holdtime}".format(**config_data["timers"])
    return command


def _tmplt_bgp_ucmp(config_data):
    command = "ucmp"
    if "fec" in config_data["ucmp"]:
        command += " fec threshold trigger"
        command += " {trigger} clear {clear} warning-only".format(**config_data["ucmp"]["fec"])
    if "link_bandwidth" in config_data["ucmp"]:
        command += " link-bandwidth {mode}".format(**config_data["ucmp"]["link_bandwidth"])
        if config_data["ucmp"]["link_bandwidth"].get("mode") == "update_delay":
            command += " {update_delay}".format(**config_data["ucmp"]["link_bandwidth"])
    if "mode" in config_data["ucmp"]:
        command += " mode 1"
        if config_data["ucmp"]["mode"].get("nexthops"):
            command += " {nexthops}".format(**config_data["ucmp"]["mode"])
    return command


def _tmplt_bgp_update(config_data):
    command = "update {wait_for}".format(**config_data["update"])
    if config_data["update"].get("batch_size"):
        command += " {batch_size}".format(**config_data["update"])
    return command


def _tmplt_bgp_vlan(config_data):
    command = "vlan {vlan}".format(**config_data)
    return command


def _tmplt_bgp_vlan_aware_bundle(config_data):
    command = "vlan-aware-bundle " + config_data["vlan_aware_bundle"]
    return command


class Bgp_globalTemplate(NetworkTemplate):
    def __init__(self, lines=None, module=None):
        super(Bgp_globalTemplate, self).__init__(
            lines=lines,
            tmplt=self,
            module=module,
        )

    # fmt: off
    PARSERS = [
        {
            "name": "router",
            "getval": re.compile(
                r"""
                ^router\s
                bgp
                \s(?P<as_num>\S+)
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_router_bgp_cmd,
            "compval": "as_number",
            "result": {"as_number": "{{ as_num }}"},
        },
        {
            "name": "vrf",
            "getval": re.compile(
                r"""
                \s*vrf
                \s(?P<vrf>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_vrf,
            "compval": "vrfs.vrf",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "vrf": "{{ vrf }}",
                    },
                },
            },
            "shared": True,
        },
        {
            "name": "aggregate_address",
            "getval": re.compile(
                r"""
                \s*aggregate-address
                \s+(?P<address>\S+)
                \s*(?P<as_set>as-set)*
                \s*(?P<summary_only>summary-only)*
                \s*(?P<attribute_map>attribute-map\s\S+)*
                \s*(?P<match_map>match-map\s\S+)*
                \s*(?P<advertise_only>advertise-only)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_aggregate_address,
            "compval": "aggregate_address",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "aggregate_address": [
                            {
                                "address": "{{ address }}",
                                "advertise_only": "{{ True if advertise_only is defined }}",
                                "as_set": "{{ True if as_set is defined }}",
                                "summary_only": "{{ True if summary_only is defined }}",
                                "attribute_map": "{{ attribute_map.split(" ")[1] }}",
                                "match_map": "{{ match_map.split(" ")[1] }}",
                            },
                        ],
                    },
                },
            },
        },
        {
            "name": "bgp_params_additional_paths",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+additional-paths
                \s+(?P<action>\S+)
                \s*(any)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.additional_paths",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "additional_paths": "{{ action }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params_advertise_inactive",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+advertise-inactive
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.advertise_inactive",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "advertise_inactive": "{{ True }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params_allowas_in",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+allowas-in
                \s*(?P<count>\d+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.allowas_in",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "allowas_in": {
                                "set": "{{ True if count is undefined }}",
                                "count": "{{ count }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params_always_compare_med",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+always-compare-med
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.always_compare_med",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "always_compare_med": "{{ True }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params_asn",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+asn
                \s+notation
                \s+(?P<notation>asdot|asplain)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.asn",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "asn": "{{ notation }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params_auto_local_addr",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+auto-local-addr
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.auto_local_addr",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "auto_local_addr": "{{ True }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params_bestpath_as_path",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+bestpath
                \s*(?P<as_path>as-path\s\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.bestpath.as_path",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "bestpath": {
                                "as_path": "{{ as_path.split(" ")[1] }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params_bestpath_ecmp_fast",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+bestpath
                \s+ecmp-fast
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.bestpath.ecmp_fast",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "bestpath": {
                                "ecmp_fast": "{{ True }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params_bestpath_med",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+bestpath
                \s+med
                \s*(?P<confed>confed)*
                \s*(?P<missing>missing-as-worst)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.bestpath.med",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "med": {
                                "confed": "{{ True if confed is defined }}",
                                "missing_as_worst": "{{ True if missing is defined }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params_bestpath_skip",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+bestpath
                \s+(?P<skip>skip)
                \s+next-hop
                \s+igp-cost
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.bestpath.skip",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "bestpath": {
                                "skip": "{{ True }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params_tie_break",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+tie-break
                \s+(?P<tie>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.bestpath.tie_break",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "tie_break": "{{ tie }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params_client_to_client",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+client-to-client
                \s+reflection
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.client_to_client",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "client_to_client": "{{ True }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params_cluster_id",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+cluster-id
                \s+(?P<address>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.cluster_id",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "cluster_id": "{{ address }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params_confederation",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+confederation
                \s*(?P<identifier>identifier\s.+)*
                \s*(?P<peers>peers\s.+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.confederation",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "confederation": {
                                "identifier": "{{ identifier.split(" ")[1] }}",
                                "peers": "{{ peers.split(" ")[1] }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params_control_plane_filter",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+control-plane-filter
                \s+default-allow
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.control_plane_filter",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "control_plane_filter": "{{ True }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params_convergence",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+convergence
                \s*(?P<slow>slow-peer)*
                \s+(?P<time>time\s\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.convergence",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "convergence": {
                                "slow_peer": "{{ True if slow is defined else False}}",
                                "time": "{{ time.split(" ")[1] }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params_default",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+default
                \s(?P<param>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.default",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "default": "{{ param }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params.enforce_first_as",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+enforce-first-as
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.enforce_first_as",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "enforce_first_as": "{{ True }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params.host_routes",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+host-routes
                \s+fib
                \s+direct-install
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.host_routes",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "host_routes": "{{ True }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params.labelled_unicast",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+labeled-unicast
                \s+rib
                \s+(?P<rib>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.labelled_unicast",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "labeled_unicast": "{{ rib }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params.listen_limit",
            "getval": re.compile(
                r"""
                \s*dynamic\speer\smax
                \s+(?P<limit>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.listen.limit",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "listen": {
                                "limit": "{{ limit }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params.listen_range",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+listen
                \s+range
                \s+(?P<address>\S+)
                \s+peer\sgroup
                \s+(?P<group>\S+)
                \s*(?P<filter>peer-filter \S+)*
                \s*(?P<remote_as>remote-as \S+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.listen.range",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "listen": {
                                "range": {
                                    "address": "{{ address }}",
                                    "peer_group": {
                                        "name": "{{ group }}",
                                        "peer_filter": "{{ filter }}",
                                        "remote_as": "{{ remote_as }}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params.log_neighbor_changes",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+log-neighbor-changes
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.log_neighbor_changes",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "log_neighbor_changes": "{{ True }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params.missing_policy",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+missing-policy
                \s+direction
                \s+(?P<dir>in|out)
                \s+action
                \s+(?P<action>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.missing_policy",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "missing_policy": {
                                "direction": "{{ dir }}",
                                "action": "{{ action }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params.monitoring",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+monitoring
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.monitoring",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "monitoring": "{{ True }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params.nexthop_unchanged",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+next-hop-unchanged
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.next_hop_unchanged",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "next_hop_unchanged": "{{ True }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params.redistribute_internal",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+redistribute-internal
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.redistribute_internal",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "redistribute_internal": "{{ True }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params.route",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+route
                \s+install-map
                \s+(?P<route>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.route",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "route": "{{ route }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params.route_reflector",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+route-reflector
                \s+preserve-attributes
                \s*(?P<preserve>always)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.route_reflector",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "route_reflector": {
                                "set": "{{ True if presever is undefined }}",
                                "preserve": "{{ True if preserve is defined }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "bgp_params.transport",
            "getval": re.compile(
                r"""
                \s*bgp
                \s+transport
                \s+listen-port
                \s+(?P<port>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_params,
            "compval": "bgp_params.transport",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bgp_params": {
                            "transport": "{{ port }}",
                        },
                    },
                },
            },
        },
        {
            "name": "default_metric",
            "getval": re.compile(
                r"""
                \s*default-metric
                \s(?P<metric>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_default_metric,
            "compval": "default_metric",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "default_metric": "{{ metric }}",
                    },
                },
            },
        },
        {
            "name": "redistribute",
            "getval": re.compile(
                r"""
                \s*redistribute
                \s+(?P<route>\S+)
                \s*(?P<level>level-1|level-2|level-1-2)*
                \s*(?P<match>match\s.+)*
                \s*(?P<route_map>route-map\s\S+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_redistribute,
            "compval": "redistribute",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "redistribute": [
                            {
                                "protocol": "{{ route }}",
                                "route_map": "{{ route_map.split(" ")[1] }}",
                                "isis_level": "{{ level }}",
                                "ospf_route": "{{ 'nssa_external_' + match.split(" ")[2] if match.split(" ")[1] == 'nssa-external' else  match.split(" ")[1]}}",
                            },
                        ],
                    },
                },
            },
        },
        {
            "name": "distance",
            "getval": re.compile(
                r"""
                \s*distance
                \s+bgp
                \s(?P<external>\d+)
                \s*(?P<internal>\d+)*
                \s*(?P<local>\d+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_distance,
            "compval": "distance",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "distance": {
                            "external": "{{ external }}",
                            "internal": "{{ internal }}",
                            "local": "{{ local }}",
                        },
                    },
                },
            },
        },
        {
            "name": "graceful_restart",
            "getval": re.compile(
                r"""
                \s*graceful-restart
                \s*(?P<restart_time>restart-time\s\d+)*
                \s*(?P<stalepath_time>stalepath-time\s\d+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_graceful_restart,
            "remval": "graceful-restart",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "graceful_restart": {
                            "set": "{{ True if restart_time and stalepath_time is not defined }}",
                            "restart_time": "{{ restart_time.split(" ")[1]|int }}",
                            "stalepath_time": "{{ stalepath_time.split(" ")[1]|int }}",
                        },
                    },
                },
            },
        },
        {
            "name": "graceful_restart_helper",
            "getval": re.compile(
                r"""
                \s*graceful-restart-helper
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_graceful_restart_helper,
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "graceful_restart": {
                            "set": "{{ True }}",
                        },
                    },
                },
            },
        },
        {
            "name": "access_group",
            "getval": re.compile(
                r"""
                \s*(?P<afi>ip|ipv6)
                \s+access-group
                \s+(?P<acl_name>\S+)
                \s*(?P<direction>in)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_access_group,
            "compval": "access_group",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "access_group": [
                            {
                                "afi": "{{ ipv4 if afi == 'ip'  else afi }}",
                                "acl_name": "{{ acl_name }}",
                                "direction": "{{ direction }}",
                            },
                        ],
                    },
                },
            },
        },
        {
            "name": "maximum_paths",
            "getval": re.compile(
                r"""
                \s*maximum-paths
                \s+(?P<max_equal_cost_paths>\d+)
                \s*(ecmp)*
                \s*(?P<max_installed_ecmp_paths>\d+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_maximum_paths,
            "compval": "maximum_paths",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "maximum_paths": {
                            "max_equal_cost_paths": "{{ max_equal_cost_paths }}",
                            "max_installed_ecmp_paths": "{{ max_installed_ecmp_paths }}",
                        },
                    },
                },
            },
        },
        {
            "name": "monitoring",
            "getval": re.compile(
                r"""
                \s*monitoring
                \s+(?P<port>\d+)
                \s*(?P<received>received\sroutes\s\S+)*
                \s*(?P<time>timestamp\s\S+)*
                \s*(?P<station>station\s\S+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_monitoring,
            "compval": "monitoring",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "monitoring": {
                            "port": "{{ port }}",
                            "received": "{{ received.split(" ")[2] }}",
                            "timestamp": "{{ timestamp.split(" ")[1] }}",
                            "station": "{{ station.split(" ")[1] }}",
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.additional_paths",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+additional-paths
                \s+(?P<action>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.additional_paths",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "additional_paths": "{{ action }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.allowas_in",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+allowas-in
                \s*(?P<count>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.allowas_in",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "allowas_in": {
                                    "set": "{{ True if count is undefined }}",
                                    "count": "{{ count }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.auto_local_addr",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+auto-local-addr
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.auto_local_addr",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "auto_local_addr": "{{ True }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.bfd",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+bfd
                \s*(?P<cbit>c-bit)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.bfd",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "bfd": "{{ 'c_bit' if cbit is defined else 'enable' }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.default_originate",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+default-originate
                \s*(?P<route_map>route-map\s\S+)*
                \s*(?P<always>always)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.default_originate",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "default_originate": {
                                    "route_map": "{{ route_map.split(" ")[1] }}",
                                    "always": "{{ True if always is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.description",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+description
                \s+(?P<desc>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.description",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "description": "{{ desc }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.dont_capability_negotiate",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+dont-capability-negotiate
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.dont_capability_negotiate",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "dont_capability_negotiate": "{{ True }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.ebgp_multihop",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+ebgp-multihop
                \s*(?P<ttl>\d+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.ebgp_multihop",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "ebgp_multihop": {
                                    "set": "{{ True if ttl is not defined }}",
                                    "ttl": "{{ ttl }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.encryption_password",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+password
                \s*(?P<type>\d+)
                \s*(?P<password>\S+)
                """,
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.encryption_password",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "peer": "{{ peer }}",
                                "encryption_password": {
                                    "type": "{{ type }}",
                                    "password": "{{ password }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.enforce_first_as",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+enforce-first-as
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.enforce_first_as",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "enforce_first_as": "{{ True }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.export_localpref",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+export-localpref
                \s+(?P<pref>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.export_localpref",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "export_localpref": "{{ pref }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.fall_over",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+fall-over
                \s+bfd
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.fall_over",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "fall_over": "{{ True }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.graceful_restart",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+graceful-restart
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "remval": "neighbor {{ peer }} graceful-restart",
            "compval": "neighbor.graceful_restart",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "graceful_restart": "{{ True }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.graceful_restart_helper",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+graceful-restart-helper
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.graceful_restart_helper",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "graceful_restart_helper": "{{ True }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.idle_restart_timer",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+idle-restart-timer
                \s+(?P<time>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.idle_restart_timer",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "idle_restart_timer": "{{ time }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.import_localpref",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+import-localpref
                \s+(?P<pref>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.import_localpref",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "import_localpref": "{{ pref }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.link_bandwidth",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+link-bandwidth
                \s*(?P<auto>auto)*
                \s*(?P<default>default\s\S+)*
                \s*(?P<update>update-delay\s\S+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.link_bandwidth",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "link_bandwidth": {
                                    "set": "{{ True }}",
                                    "auto": "{{ True if auto is defined }}",
                                    "default": "{{ default.split(" ")[1] if default is defined }}",
                                    "update_delay": "{{ update.split(" ")[1] if update is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.local_as",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+local-as
                \s+(?P<num>\S+)
                \s+no-prepend
                \s+replace-as
                \s*(?P<fallback>fallback)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.local_as",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "local_as": {
                                    "as_number": "{{ num }}",
                                    "fallback": "{{ True if fallback is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.local_v6_addr",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+local-v6-addr
                \s+(?P<addr>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.local_v6_addr",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "local_v6_addr": "{{ addr }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.maximum_accepted_routes",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+maximum-accepted-routes
                \s+(?P<count>\d+)
                \s*warning-limit*
                \s*(?P<limit>\d+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.maximum_accepted_routes",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "maximum_accepted_routes": {
                                    "count": "{{ count }}",
                                    "warning_limit": "{{ limit }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.maximum_received_routes",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+maximum-routes
                \s+(?P<count>\d+)*
                \s*(warning-limit)*
                \s*(?P<limit>\d+)*
                \s*(?P<percent>percent)*
                \s*(?P<warning_only>warning-only)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.maximum_received_routes",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "maximum_received_routes": {
                                    "count": "{{ count }}",
                                    "warning_limit": {
                                        "limit_count": "{{ limit if percent is undefined }}",
                                        "limit_percent": "{{ limit if percent is defined }}",
                                    },
                                    "warning_only": "{{ True if warning_only is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.metric_out",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+metric-out
                \s+(?P<metric>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.metric_out",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "metric_out": "{{ metric }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.monitoring",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+monitoring
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.monitoring",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "monitoring": "{{ True }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.next_hop_self",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+next-hop-self
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.next_hop_self",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "next_hop_self": "{{ True }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.next_hop_unchanged",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+next-hop-unchanged
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.next_hop_unchanged",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "next_hop_unchanged": "{{ True }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.next_hop_v6_addr",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+next-hop-v6-addr
                \s+(?P<addr>\S+)
                \s+in
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.next_hop_v6_address",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "next_hop_v6_address": "{{ addr }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.out_delay",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+out-delay
                \s+(?P<delay>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.out_delay",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "out_delay": "{{ delay }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.remote_as",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+remote-as
                \s+(?P<num>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.remote_as",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "remote_as": "{{ num }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.remove_private_as",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+remove-private-as
                \s*(?P<all>all)*
                \s*(?P<replace>replace-as)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.remove_private_as",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "remove_private_as": {
                                    "set": "{{ True if all is undefined and replace is undefined }}",
                                    "all": "{{ True if all is defined }}",
                                    "replace_as": "{{ True if replace is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.peer_group",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+peer\sgroup
                \s*(?P<name>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.peer_group",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "peer_group": "{{ name if name is defined else peer}}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.prefix_list",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+prefix-list
                \s+(?P<name>\S+)
                \s+(?P<dir>in|out)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.prefix_list",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "prefix_list": {
                                    "name": "{{ name }}",
                                    "direction": "{{ dir }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.route_map",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+route-map
                \s+(?P<name>\S+)
                \s+(?P<dir>in|out)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.route_map",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "route_map": {
                                    "name": "{{ name }}",
                                    "direction": "{{ dir }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.route_reflector_client",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+route-reflector-client
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.route_reflector_client",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "route_reflector_client": "{{ True }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.route_to_peer",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+route-to-peer
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.route_to_peer",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "route_to_peer": "{{ True }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.send_community",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+send-community
                \s*(?P<comm>add|extended|link-bandwidth|remove|standard)*
                \s*(?P<attr>extended|link-bandwidth|standard)*
                \s*(?P<link>aggregate|divide)*
                \s*(?P<div>equal|ratio)*
                \s*(?P<speed>\S+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.send_community",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "send_community": {
                                    "set": "{{ True if comm is not defined }}",
                                    "community_attribute": "{{ comm }}",
                                    "sub_attribute": "{{ attr }}",
                                    "link_bandwidth_attribute": "{{ link }}",
                                    "speed": "{{ speed }}",
                                    "divide": "{{ div }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.shutdown",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+shutdown
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.shutdown",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "shutdown": "{{ True }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.soft_reconfiguration",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+soft-reconfiguration
                \s+inbound
                \s*(?P<all>\S+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.soft_reconfiguration",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "soft_reconfiguration": "{{ 'all' if all is defined else 'None' }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.transport",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+transport
                \s+(?P<mode>\S+)
                \s*(passive)*
                \s*(?P<port>\d+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.transport",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "transport": {
                                    "connection_mode": "{{ mode }}",
                                    "remote_port": "{{ port if port is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.timers",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+timers
                \s+(?P<keepalive>\d+)
                \s+(?P<hold>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.timers",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "timers": {
                                    "keepalive": "{{ keepalive }}",
                                    "holdtime": "{{ holdtime }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.ttl",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+ttl
                \s+maximum-hops
                \s+(?P<hop>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.ttl",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "ttl": "{{ hop }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.update_source",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+update-source
                \s+(?P<src>.+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.update_source",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "update_source": "{{ src }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor.weight",
            "getval": re.compile(
                r"""
                \s*neighbor
                \s+(?P<peer>\S+)
                \s+weight
                \s+(?P<val>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_neighbor,
            "compval": "neighbor.weight",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "neighbor": {
                            "{{ peer }}": {
                                "neighbor_address": "{{ peer }}",
                                "weight": "{{ val }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "network",
            "getval": re.compile(
                r"""
                \s*network
                \s+(?P<address>\S+)
                \s*(route-map)*
                \s*(?P<route_map>\S+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_network,
            "compval": "network",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "network": {
                            "{{ address }}": {
                                "address": "{{ address }}",
                                "route_map": "{{ route_map }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "route_target",
            "getval": re.compile(
                r"""
                \s*route-target
                \s+(?P<action>\S+)
                \s*(?P<type>evpn|vpn-ipv4|vpn-ipv6)*
                \s*(?P<map>route-map\s\S+)*
                \s*(?P<imp>imported-route)*
                \s*(?P<target>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_route_target,
            "compval": "route_target",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "route_target": {
                            "action": "{{ action }}",
                            "type": "{{ type }}",
                            "route_map": "{{ map.split(" ")[1] }}",
                            "imported_route": "{{ True if imp is defined }}",
                            "target": "{{ target }}",
                        },
                    },
                },
            },
        },
        {
            "name": "router_id",
            "getval": re.compile(
                r"""
                \s*router-id
                \s+(?P<id>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_router_id,
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "router_id": "{{ id }}",
                    },
                },
            },
        },
        {
            "name": "shutdown",
            "getval": re.compile(
                r"""
                \s*shutdown
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_shutdown,
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "shutdown": "{{ True }}",
                    },
                },
            },
        },
        {
            "name": "timers",
            "getval": re.compile(
                r"""
                \s*timers
                \s+bgp
                \s+(?P<keepalive>\d+)
                \s+(?P<holdtime>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_timers,
            "compval": "timers",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "timers": {
                            "keepalive": "{{ keepalive }}",
                            "holdtime": "{{ holdtime }}",
                        },
                    },
                },
            },
        },
        {
            "name": "ucmp_fec",
            "getval": re.compile(
                r"""
                \s*ucmp
                \s+fec
                \s+threshold
                \s+(?P<trigger>trigger \d+)
                \s+(?P<clear>clear \d+)
                \s+warning-only
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_ucmp,
            "compval": "ucmp.fec",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "ucmp": {
                            "fec": {
                                "trigger": "{{ trigger }}",
                                "clear": "{{ clear }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ucmp_link_bandwidth",
            "getval": re.compile(
                r"""
                \s*ucmp
                \s+link-bandwidth
                \s*(?P<ucmp_mode>recursive|encoding-weighted|update-delay)
                \s*(?P<update_delay>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_ucmp,
            "compval": "ucmp.link_bandwidth",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "ucmp": {
                            "link_bandwidth": {
                                "mode": "{{ ucmp_mode }}",
                                "update_delay": "{{ update_delay }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ucmp_mode",
            "getval": re.compile(
                r"""
                \s*ucmp
                \s+mode
                \s+(?P<ucmp_set>\d+)
                \s*(?P<nexthop>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_ucmp,
            "compval": "ucmp.mode",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "ucmp": {
                            "mode": {
                                "set": "{{ True if ucmp_set == '1'}}",
                                "nexthops": "{{ nexthop }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "update",
            "getval": re.compile(
                r"""
                \s*update
                \s+(?P<wait>\S+)
                \s*(?P<size>batch-size\s\d+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_update,
            "compval": "update",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "update": {
                            "wait_for": "{{ wait }}",
                            "batch_size": "{{ size.split(" ")[1] }}",
                        },
                    },
                },
            },
        },
        {
            "name": "vlan",
            "getval": re.compile(
                r"""
                \s*vlan
                \s+(?P<id>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_vlan,
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "vlan": "{{ id }}",
                    },
                },
            },
        },
        {
            "name": "vlan_aware_bundle",
            "getval": re.compile(
                r"""
                \s*vlan-aware-bundle
                \s+(?P<bundle>.+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_bgp_vlan_aware_bundle,
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "vlan_aware_bundle": "{{ bundle }}",
                    },
                },
            },
        },
    ]
    # fmt: on



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2024, David Neilan (@dneilan-intel) <david.neilan@intel.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {
    "metadata_version": "1.1",
    "status": ["preview"],
    "supported_by": "certified",
}

DOCUMENTATION = r"""
---
module: aci_system_connectivity_preference
short_description: APIC Connectivity Preferences (mgmt:ConnectivityPrefs)
description:
- Manages APIC Connectivity Preferences on Cisco ACI fabrics.
options:
  interface_preference:
    description:
    - Interface to use for external connection.
    type: str
    choices: [ inband, ooband ]
    aliases: [ interface_pref, int_pref, external_connection ]

  state:
    description:
    - Use C(present) for updating.
    - Use C(query) for listing an object.
    type: str
    choices: [ present, query ]
    default: present

extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation
- cisco.aci.owner

seealso:
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(mgmt:ConnectivityPrefs).
  link: https://developer.cisco.com/docs/apic-mim-ref/

author:
- David Neilan (@dneilan-intel)
"""

EXAMPLES = r"""
- name: Configure Out-of-band Connectivity Preference
  cisco.aci.aci_system_connectivity_preference:
    host: apic
    username: admin
    password: SomeSecretPassword
    interface_preference: ooband
    state: present
  delegate_to: localhost

- name: Configure In-band Connectivity Preference
  cisco.aci.aci_system_connectivity_preference:
    host: apic
    username: admin
    password: SomeSecretPassword
    interface_preference: inband
    state: present
  delegate_to: localhost

- name: Query Management Connectivity Preference
  cisco.aci.aci_system_connectivity_preference:
    host: apic
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost
"""

RETURN = r"""
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "mgmtConnectivityPrefs": {
                "attributes": {
                    "annotation": "",
                    "childAction": "",
                    "descr": "",
                    "dn": "uni/fabric/connectivityPrefs",
                    "extMngdBy": "",
                    "interfacePref": "inband",
                    "lcOwn": "local",
                    "modTs": "2023-11-14T16:25:32.629+00:00",
                    "name": "default",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": "",
                    "status": "",
                    "uid": "0"
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
      "code": "120",
      "text": "unknown property value foo, name interfacePref, class mgmtConnectivityPrefs [(Dn0)] Dn0=, "
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: str
  sample:
    {
      "totalCount": "1",
      "imdata": [
          {
              "error": {
                  "attributes": {
                      "code": "120",
                      "text": "unknown property value foo, name interfacePref, class mgmtConnectivityPrefs [(Dn0)] Dn0=, "
                  }
              }
          }
      ]
    }
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "mgmtConnectivityPrefs": {
            "attributes": {
                "interfacePref": "inband"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "mgmtConnectivityPrefs": {
                "attributes": {
                    "annotation": "",
                    "descr": "",
                    "dn": "uni/fabric/connectivityPrefs",
                    "interfacePref": "ooband",
                    "name": "default",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "mgmtConnectivityPrefs": {
            "attributes": {
                "interfacePref": "inband"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: str
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: str
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: str
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: str
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import (
    ACIModule,
    aci_argument_spec,
    aci_annotation_spec,
    aci_owner_spec,
)


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(aci_owner_spec())
    argument_spec.update(
        interface_preference=dict(
            type="str",
            choices=["ooband", "inband"],
            aliases=["interface_pref", "int_pref", "external_connection"],
        ),
        state=dict(type="str", default="present", choices=["present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "present", ["interface_preference"]],
        ],
    )
    aci = ACIModule(module)

    interface_preference = module.params.get("interface_preference")
    state = module.params.get("state")

    aci.construct_url(
        root_class=dict(
            aci_class="mgmtConnectivityPrefs",
            aci_rn="fabric/connectivityPrefs",
        ),
    )
    aci.get_existing()

    if state == "present":
        aci.payload(
            aci_class="mgmtConnectivityPrefs",
            class_config=dict(interfacePref=interface_preference),
        )

        aci.get_diff(aci_class="mgmtConnectivityPrefs")

        aci.post_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: interface_update_v1
short_description: Resource module for Interface Update V1
description:
- Manage operation update of the resource Interface Update V1.
- >
   Add/Update Interface description, VLAN membership, Voice VLAN and change Interface admin status 'UP'/'DOWN' from
   Request body.
version_added: '6.0.0'
extends_documentation_fragment:
  - cisco.dnac.module
author: Rafael Campos (@racampos)
options:
  adminStatus:
    description: Admin status as ('UP'/'DOWN').
    type: str
  deploymentMode:
    description: DeploymentMode query parameter. Preview/Deploy 'Preview' means the
      configuration is not pushed to the device. 'Deploy' makes the configuration pushed
      to the device.
    type: str
  description:
    description: Description for the Interface.
    type: str
  interfaceUuid:
    description: InterfaceUuid path parameter. Interface ID.
    type: str
  vlanId:
    description: VLAN Id to be Updated.
    type: int
  voiceVlanId:
    description: Voice Vlan Id to be Updated.
    type: int
requirements:
- dnacentersdk >= 2.4.9
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Devices UpdateInterfaceDetailsV1
  description: Complete reference of the UpdateInterfaceDetailsV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!update-interface-details
notes:
  - SDK Method used are
    devices.Devices.update_interface_details_v1,

  - Paths used are
    put /dna/intent/api/v1/interface/{interfaceUuid},

"""

EXAMPLES = r"""
- name: Update by id
  cisco.dnac.interface_update_v1:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    adminStatus: string
    deploymentMode: string
    description: string
    interfaceUuid: string
    vlanId: 0
    voiceVlanId: 0

"""
RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": {
        "type": "string",
        "properties": {
          "taskId": {
            "type": "string"
          },
          "url": {
            "type": "string"
          }
        },
        "required": [
          "string"
        ]
      },
      "version": {
        "type": "string"
      }
    }
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)


DOCUMENTATION = r"""
---
module: sda_fabric_devices_layer2_handoffs
short_description: Resource module for Sda Fabric Devices Layer2 Handoffs
description:
- This module represents an alias of the module sda_fabric_devices_layer2_handoffs_v1
version_added: '6.14.0'
extends_documentation_fragment:
  - cisco.dnac.module
author: Rafael Campos (@racampos)
options:
  fabricId:
    description: FabricId query parameter. ID of the fabric this device belongs to.
    type: str
  id:
    description: Id path parameter. ID of the layer 2 handoff of a fabric device.
    type: str
  networkDeviceId:
    description: NetworkDeviceId query parameter. Network device ID of the fabric device.
    type: str
  payload:
    description: Sda Fabric Devices Layer2 Handoffs's payload.
    elements: dict
    suboptions:
      externalVlanId:
        description: External VLAN number into which the fabric must be extended. Allowed
          VLAN range is 2-4094 except for reserved vlans (1, 1002-1005, 2046, 4094).
        type: int
      fabricId:
        description: ID of the fabric this device is assigned to.
        type: str
      interfaceName:
        description: Interface name of the layer 2 handoff. E.g., GigabitEthernet1/0/4.
        type: str
      internalVlanId:
        description: VLAN number associated with this fabric. Allowed VLAN range is
          2-4094 except for reserved vlans (1, 1002-1005, 2046, 4094).
        type: int
      networkDeviceId:
        description: Network device ID of the fabric device.
        type: str
    type: list
requirements:
- dnacentersdk >= 2.4.9
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for SDA AddFabricDevicesLayer2HandoffsV1
  description: Complete reference of the AddFabricDevicesLayer2HandoffsV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!add-fabric-devices-layer-2-handoffs
- name: Cisco DNA Center documentation for SDA DeleteFabricDeviceLayer2HandoffByIdV1
  description: Complete reference of the DeleteFabricDeviceLayer2HandoffByIdV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!delete-fabric-device-layer-2-handoff-by-id
- name: Cisco DNA Center documentation for SDA DeleteFabricDeviceLayer2HandoffsV1
  description: Complete reference of the DeleteFabricDeviceLayer2HandoffsV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!delete-fabric-device-layer-2-handoffs
notes:
  - SDK Method used are
    sda.Sda.add_fabric_devices_layer2_handoffs_v1,
    sda.Sda.delete_fabric_device_layer2_handoff_by_id_v1,

  - Paths used are
    post /dna/intent/api/v1/sda/fabricDevices/layer2Handoffs,
    delete /dna/intent/api/v1/sda/fabricDevices/layer2Handoffs,
    delete /dna/intent/api/v1/sda/fabricDevices/layer2Handoffs/{id},
  - It should be noted that this module is an alias of sda_fabric_devices_layer2_handoffs_v1

"""

EXAMPLES = r"""
- name: Delete all
  cisco.dnac.sda_fabric_devices_layer2_handoffs:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    state: absent
    fabricId: string
    networkDeviceId: string

- name: Create
  cisco.dnac.sda_fabric_devices_layer2_handoffs:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    state: present
    payload:
    - externalVlanId: 0
      fabricId: string
      interfaceName: string
      internalVlanId: 0
      networkDeviceId: string

- name: Delete by id
  cisco.dnac.sda_fabric_devices_layer2_handoffs:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    state: absent
    id: string

"""
RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": {
        "taskId": "string",
        "url": "string"
      },
      "version": "string"
    }
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: assurance_issues_update_v1
short_description: Resource module for Assurance Issues Update V1
description:
- Manage operation create of the resource Assurance Issues Update V1.
- >
   Updates selected fields in the given issue. Currently the only field that can be updated is 'notes' field. For
   detailed information about the usage of the API, please refer to the Open API specification document - https
   //github.com/cisco-en-programmability/catalyst-center-api-specs/blob/main/Assurance/CE_Cat_Center_Org-
   IssuesLifecycle-1.0.0-resolved.yaml.
version_added: '6.15.0'
extends_documentation_fragment:
  - cisco.dnac.module
author: Rafael Campos (@racampos)
options:
  headers:
    description: Additional headers.
    type: dict
  id:
    description: Id path parameter. The issue Uuid.
    type: str
  notes:
    description: Notes.
    type: str
requirements:
- dnacentersdk >= 2.4.9
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Issues UpdateTheGivenIssueByUpdatingSelectedFieldsV1
  description: Complete reference of the UpdateTheGivenIssueByUpdatingSelectedFieldsV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!update-the-given-issue-by-updating-selected-fields
notes:
  - SDK Method used are
    issues.Issues.update_the_given_issue_by_updating_selected_fields_v1,

  - Paths used are
    post /dna/intent/api/v1/assuranceIssues/{id}/update,

"""

EXAMPLES = r"""
- name: Create
  cisco.dnac.assurance_issues_update_v1:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    headers: '{{my_headers | from_json}}'
    id: string
    notes: string

"""
RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": {
        "issueId": "string",
        "name": "string",
        "description": "string",
        "summary": "string",
        "priority": "string",
        "severity": "string",
        "deviceType": "string",
        "category": "string",
        "entityType": "string",
        "entityId": "string",
        "firstOccurredTime": 0,
        "mostRecentOccurredTime": 0,
        "status": "string",
        "isGlobal": true,
        "updatedBy": "string",
        "updatedTime": 0,
        "notes": "string",
        "siteId": "string",
        "siteHierarchyId": "string",
        "siteName": "string",
        "siteHierarchy": "string",
        "suggestedActions": [
          {
            "message": "string"
          }
        ],
        "additionalAttributes": [
          {
            "key": "string",
            "value": "string"
          }
        ]
      },
      "version": "string"
    }
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: ec2_metadata_facts
version_added: 1.0.0
short_description: Gathers facts (instance metadata) about remote hosts within EC2
author:
    - Silviu Dicu (@silviud)
    - Vinay Dandekar (@roadmapper)
description:
    - This module fetches data from the instance metadata endpoint in EC2 as per
      U(https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/ec2-instance-metadata.html).
    - The module must be called from within the EC2 instance itself.
    - The module is configured to utilize the session oriented Instance Metadata Service v2 (IMDSv2)
      U(https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/configuring-instance-metadata-service.html).
    - If the HttpEndpoint parameter
      U(https://docs.aws.amazon.com/AWSEC2/latest/APIReference/API_ModifyInstanceMetadataOptions.html#API_ModifyInstanceMetadataOptions_RequestParameters)
      is set to disabled for the EC2 instance, the module will return an error while retrieving a session token.
notes:
    - Parameters to filter on ec2_metadata_facts may be added later.
options:
    metadata_token_ttl_seconds:
        description:
            - Specify a value for the C(X-aws-ec2-metadata-token-ttl-seconds) header.
            - Value must be between V(1) and V(21600).
        type: int
        default: 60
        version_added: 8.2.0
"""

EXAMPLES = r"""
# Gather EC2 metadata facts
- amazon.aws.ec2_metadata_facts:

# Set a bigger value for X-aws-ec2-metadata-token-ttl-seconds header
- amazon.aws.ec2_metadata_facts:
    metadata_token_ttl_seconds: 240

- debug:
    msg: "This instance is a t1.micro"
  when: ansible_ec2_instance_type == "t1.micro"
"""

RETURN = r"""
ansible_facts:
    description: Dictionary of new facts representing discovered properties of the EC2 instance.
    returned: changed
    type: complex
    contains:
        ansible_ec2_ami_id:
            description: The AMI ID used to launch the instance.
            type: str
            sample: "ami-XXXXXXXX"
        ansible_ec2_ami_launch_index:
            description:
                - If you started more than one instance at the same time, this value indicates the order in which the instance was launched.
                - The value of the first instance launched is 0.
            type: str
            sample: "0"
        ansible_ec2_ami_manifest_path:
            description:
                - The path to the AMI manifest file in Amazon S3.
                - If you used an Amazon EBS-backed AMI to launch the instance, the returned result is unknown.
            type: str
            sample: "(unknown)"
        ansible_ec2_ancestor_ami_ids:
            description:
                - The AMI IDs of any instances that were rebundled to create this AMI.
                - This value will only exist if the AMI manifest file contained an ancestor-amis key.
            type: str
            sample: "(unknown)"
        ansible_ec2_block_device_mapping_ami:
            description: The virtual device that contains the root/boot file system.
            type: str
            sample: "/dev/sda1"
        ansible_ec2_block_device_mapping_ebsN:
            description:
                - The virtual devices associated with Amazon EBS volumes, if any are present.
                - Amazon EBS volumes are only available in metadata if they were present at launch time or when the instance was last started.
                - The N indicates the index of the Amazon EBS volume (such as ebs1 or ebs2).
            type: str
            sample: "/dev/xvdb"
        ansible_ec2_block_device_mapping_ephemeralN:
            description: The virtual devices associated with ephemeral devices, if any are present. The N indicates the index of the ephemeral volume.
            type: str
            sample: "/dev/xvdc"
        ansible_ec2_block_device_mapping_root:
            description:
                - The virtual devices or partitions associated with the root devices, or partitions on the virtual device,
                  where the root (/ or C) file system is associated with the given instance.
            type: str
            sample: "/dev/sda1"
        ansible_ec2_block_device_mapping_swap:
            description: The virtual devices associated with swap. Not always present.
            type: str
            sample: "/dev/sda2"
        ansible_ec2_fws_instance_monitoring:
            description: "Value showing whether the customer has enabled detailed one-minute monitoring in CloudWatch."
            type: str
            sample: "enabled"
        ansible_ec2_hostname:
            description:
                - The private IPv4 DNS hostname of the instance.
                - In cases where multiple network interfaces are present, this refers to the eth0 device (the device for which the device number is 0).
            type: str
            sample: "ip-10-0-0-1.ec2.internal"
        ansible_ec2_iam_info:
            description:
                - If there is an IAM role associated with the instance, contains information about the last time the instance profile was updated,
                  including the instance's LastUpdated date, InstanceProfileArn, and InstanceProfileId. Otherwise, not present.
            type: complex
            sample: ""
            contains:
                LastUpdated:
                    description: The last time which InstanceProfile is associated with the Instance changed.
                    type: str
                InstanceProfileArn:
                    description: The ARN of the InstanceProfile associated with the Instance.
                    type: str
                InstanceProfileId:
                    description: The Id of the InstanceProfile associated with the Instance.
                    type: str
        ansible_ec2_iam_info_instanceprofilearn:
            description: The IAM instance profile ARN.
            type: str
            sample: "arn:aws:iam::123456789012:instance-profile/role_name"
        ansible_ec2_iam_info_instanceprofileid:
            description: IAM instance profile ID.
            type: str
            sample: ""
        ansible_ec2_iam_info_lastupdated:
            description: IAM info last updated time.
            type: str
            sample: "2017-05-12T02:42:27Z"
        ansible_ec2_iam_instance_profile_role:
            description: IAM instance role.
            type: str
            sample: "role_name"
        ansible_ec2_iam_security_credentials_role_name:
            description:
                - If there is an IAM role associated with the instance, role-name is the name of the role,
                  and role-name contains the temporary security credentials associated with the role. Otherwise, not present.
            type: str
            sample: ""
        ansible_ec2_iam_security_credentials_role_name_accesskeyid:
            description: IAM role access key ID.
            type: str
            sample: ""
        ansible_ec2_iam_security_credentials_role_name_code:
            description: IAM code.
            type: str
            sample: "Success"
        ansible_ec2_iam_security_credentials_role_name_expiration:
            description: IAM role credentials expiration time.
            type: str
            sample: "2017-05-12T09:11:41Z"
        ansible_ec2_iam_security_credentials_role_name_lastupdated:
            description: IAM role last updated time.
            type: str
            sample: "2017-05-12T02:40:44Z"
        ansible_ec2_iam_security_credentials_role_name_secretaccesskey:
            description: IAM role secret access key.
            type: str
            sample: ""
        ansible_ec2_iam_security_credentials_role_name_token:
            description: IAM role token.
            type: str
            sample: ""
        ansible_ec2_iam_security_credentials_role_name_type:
            description: IAM role type.
            type: str
            sample: "AWS-HMAC"
        ansible_ec2_instance_action:
            description: Notifies the instance that it should reboot in preparation for bundling.
            type: str
            sample: "none"
        ansible_ec2_instance_id:
            description: The ID of this instance.
            type: str
            sample: "i-XXXXXXXXXXXXXXXXX"
        ansible_ec2_instance_identity_document:
            description: JSON containing instance attributes, such as instance-id, private IP address, etc.
            type: str
            sample: ""
        ansible_ec2_instance_identity_document_accountid:
            description: ""
            type: str
            sample: "123456789012"
        ansible_ec2_instance_identity_document_architecture:
            description: Instance system architecture.
            type: str
            sample: "x86_64"
        ansible_ec2_instance_identity_document_availabilityzone:
            description: The Availability Zone in which the instance launched.
            type: str
            sample: "us-east-1a"
        ansible_ec2_instance_identity_document_billingproducts:
            description: Billing products for this instance.
            type: str
            sample: ""
        ansible_ec2_instance_identity_document_devpayproductcodes:
            description: Product codes for the launched AMI.
            type: str
            sample: ""
        ansible_ec2_instance_identity_document_imageid:
            description: The AMI ID used to launch the instance.
            type: str
            sample: "ami-01234567"
        ansible_ec2_instance_identity_document_instanceid:
            description: The ID of this instance.
            type: str
            sample: "i-0123456789abcdef0"
        ansible_ec2_instance_identity_document_instancetype:
            description: The type of instance.
            type: str
            sample: "m4.large"
        ansible_ec2_instance_identity_document_kernelid:
            description: The ID of the kernel launched with this instance, if applicable.
            type: str
            sample: ""
        ansible_ec2_instance_identity_document_pendingtime:
            description: The instance pending time.
            type: str
            sample: "2017-05-11T20:51:20Z"
        ansible_ec2_instance_identity_document_privateip:
            description:
                - The private IPv4 address of the instance.
                - In cases where multiple network interfaces are present, this refers to the eth0 device (the device for which the device number is 0).
            type: str
            sample: "10.0.0.1"
        ansible_ec2_instance_identity_document_ramdiskid:
            description: The ID of the RAM disk specified at launch time, if applicable.
            type: str
            sample: ""
        ansible_ec2_instance_identity_document_region:
            description: The Region in which the instance launched.
            type: str
            sample: "us-east-1"
        ansible_ec2_instance_identity_document_version:
            description: Identity document version.
            type: str
            sample: "2010-08-31"
        ansible_ec2_instance_identity_pkcs7:
            description: Used to verify the document's authenticity and content against the signature.
            type: str
            sample: ""
        ansible_ec2_instance_identity_rsa2048:
            description: Used to verify the document's authenticity and content against the signature.
            type: str
            sample: ""
        ansible_ec2_instance_identity_signature:
            description: Data that can be used by other parties to verify its origin and authenticity.
            type: str
            sample: ""
        ansible_ec2_instance_life_cycle:
            description: The purchasing option of the instance.
            type: str
            sample: "on-demand"
        ansible_ec2_instance_tags_keys:
            description:
                - The list of tags keys of the instance.
                - Returns empty list if access to tags (InstanceMetadataTags) in instance metadata is not enabled.
            type: list
            elements: str
            sample: ["tagKey1", "tag_key2"]
            version_added: 5.5.0
        ansible_ec2_instance_type:
            description: The type of the instance.
            type: str
            sample: "m4.large"
        ansible_ec2_local_hostname:
            description:
                - The private IPv4 DNS hostname of the instance.
                - In cases where multiple network interfaces are present, this refers to the eth0 device (the device for which the device number is 0).
            type: str
            sample: "ip-10-0-0-1.ec2.internal"
        ansible_ec2_local_ipv4:
            description:
                - The private IPv4 address of the instance.
                - In cases where multiple network interfaces are present, this refers to the eth0 device (the device for which the device number is 0).
            type: str
            sample: "10.0.0.1"
        ansible_ec2_mac:
            description:
                - The instance's media access control (MAC) address.
                - In cases where multiple network interfaces are present, this refers to the eth0 device (the device for which the device number is 0).
            type: str
            sample: "00:11:22:33:44:55"
        ansible_ec2_metrics_vhostmd:
            description: Metrics; no longer available.
            type: str
            sample: ""
        ansible_ec2_network_interfaces_macs_mac_address_device_number:
            description:
                - The unique device number associated with that interface. The device number corresponds to the device name;
                  for example, a device-number of 2 is for the eth2 device.
                - This category corresponds to the DeviceIndex and device-index fields that are used by the Amazon EC2 API and the EC2 commands for the AWS CLI.
            type: str
            sample: "0"
        ansible_ec2_network_interfaces_macs_mac_address_interface_id:
            description: The elastic network interface ID.
            type: str
            sample: "eni-12345678"
        ansible_ec2_network_interfaces_macs_mac_address_ipv4_associations_ip_address:
            description: The private IPv4 addresses that are associated with each public-ip address and assigned to that interface.
            type: str
            sample: ""
        ansible_ec2_network_interfaces_macs_mac_address_ipv6s:
            description: The IPv6 addresses associated with the interface. Returned only for instances launched into a VPC.
            type: str
            sample: ""
        ansible_ec2_network_interfaces_macs_mac_address_local_hostname:
            description: The interface's local hostname.
            type: str
            sample: ""
        ansible_ec2_network_interfaces_macs_mac_address_local_ipv4s:
            description: The private IPv4 addresses associated with the interface.
            type: str
            sample: ""
        ansible_ec2_network_interfaces_macs_mac_address_mac:
            description: The instance's MAC address.
            type: str
            sample: "00:11:22:33:44:55"
        ansible_ec2_network_interfaces_macs_mac_address_owner_id:
            description:
                - The ID of the owner of the network interface.
                - In multiple-interface environments, an interface can be attached by a third party, such as Elastic Load Balancing.
                - Traffic on an interface is always billed to the interface owner.
            type: str
            sample: "123456789012"
        ansible_ec2_network_interfaces_macs_mac_address_public_hostname:
            description:
                - The interface's public DNS (IPv4). If the instance is in a VPC,
                  this category is only returned if the enableDnsHostnames attribute is set to true.
            type: str
            sample: "ec2-1-2-3-4.compute-1.amazonaws.com"
        ansible_ec2_network_interfaces_macs_mac_address_public_ipv4s:
            description: The Elastic IP addresses associated with the interface. There may be multiple IPv4 addresses on an instance.
            type: str
            sample: "1.2.3.4"
        ansible_ec2_network_interfaces_macs_mac_address_security_group_ids:
            description: The IDs of the security groups to which the network interface belongs. Returned only for instances launched into a VPC.
            type: str
            sample: "sg-01234567,sg-01234568"
        ansible_ec2_network_interfaces_macs_mac_address_security_groups:
            description: Security groups to which the network interface belongs. Returned only for instances launched into a VPC.
            type: str
            sample: "secgroup1,secgroup2"
        ansible_ec2_network_interfaces_macs_mac_address_subnet_id:
            description: The ID of the subnet in which the interface resides. Returned only for instances launched into a VPC.
            type: str
            sample: "subnet-01234567"
        ansible_ec2_network_interfaces_macs_mac_address_subnet_ipv4_cidr_block:
            description: The IPv4 CIDR block of the subnet in which the interface resides. Returned only for instances launched into a VPC.
            type: str
            sample: "10.0.1.0/24"
        ansible_ec2_network_interfaces_macs_mac_address_subnet_ipv6_cidr_blocks:
            description: The IPv6 CIDR block of the subnet in which the interface resides. Returned only for instances launched into a VPC.
            type: str
            sample: ""
        ansible_ec2_network_interfaces_macs_mac_address_vpc_id:
            description: The ID of the VPC in which the interface resides. Returned only for instances launched into a VPC.
            type: str
            sample: "vpc-0123456"
        ansible_ec2_network_interfaces_macs_mac_address_vpc_ipv4_cidr_block:
            description: The IPv4 CIDR block of the VPC in which the interface resides. Returned only for instances launched into a VPC.
            type: str
            sample: "10.0.0.0/16"
        ansible_ec2_network_interfaces_macs_mac_address_vpc_ipv4_cidr_blocks:
            description: The IPv4 CIDR block of the VPC in which the interface resides. Returned only for instances launched into a VPC.
            type: str
            sample: "10.0.0.0/16"
        ansible_ec2_network_interfaces_macs_mac_address_vpc_ipv6_cidr_blocks:
            description: The IPv6 CIDR block of the VPC in which the interface resides. Returned only for instances launched into a VPC.
            type: str
            sample: ""
        ansible_ec2_placement_availability_zone:
            description: The Availability Zone in which the instance launched.
            type: str
            sample: "us-east-1a"
        ansible_ec2_placement_region:
            description: The Region in which the instance launched.
            type: str
            sample: "us-east-1"
        ansible_ec2_product_codes:
            description: Product codes associated with the instance, if any.
            type: str
            sample: "aw0evgkw8e5c1q413zgy5pjce"
        ansible_ec2_profile:
            description: EC2 instance hardware profile.
            type: str
            sample: "default-hvm"
        ansible_ec2_public_hostname:
            description:
                - The instance's public DNS. If the instance is in a VPC, this category is only returned if the enableDnsHostnames attribute is set to true.
            type: str
            sample: "ec2-1-2-3-4.compute-1.amazonaws.com"
        ansible_ec2_public_ipv4:
            description: The public IPv4 address. If an Elastic IP address is associated with the instance, the value returned is the Elastic IP address.
            type: str
            sample: "1.2.3.4"
        ansible_ec2_public_key:
            description: Public key. Only available if supplied at instance launch time.
            type: str
            sample: ""
        ansible_ec2_ramdisk_id:
            description: The ID of the RAM disk specified at launch time, if applicable.
            type: str
            sample: ""
        ansible_ec2_reservation_id:
            description: The ID of the reservation.
            type: str
            sample: "r-0123456789abcdef0"
        ansible_ec2_security_groups:
            description:
                - The names of the security groups applied to the instance. After launch, you can only change the security groups of instances running in a VPC.
                - Such changes are reflected here and in network/interfaces/macs/mac/security-groups.
            type: str
            sample: "securitygroup1,securitygroup2"
        ansible_ec2_services_domain:
            description: The domain for AWS resources for the region; for example, amazonaws.com for us-east-1.
            type: str
            sample: "amazonaws.com"
        ansible_ec2_services_partition:
            description:
                - The partition that the resource is in. For standard AWS regions, the partition is aws.
                - If you have resources in other partitions, the partition is aws-partitionname.
                - For example, the partition for resources in the China (Beijing) region is aws-cn.
            type: str
            sample: "aws"
        ansible_ec2_spot_termination_time:
            description:
                - The approximate time, in UTC, that the operating system for your Spot instance will receive the shutdown signal.
                - This item is present and contains a time value only if the Spot instance has been marked for termination by Amazon EC2.
                - The termination-time item is not set to a time if you terminated the Spot instance yourself.
            type: str
            sample: "2015-01-05T18:02:00Z"
        ansible_ec2_user_data:
            description: The instance user data.
            type: str
            sample: "#!/bin/bash"
"""

import json
import re
import socket
import time
import zlib

from ansible.module_utils._text import to_text
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.six.moves.urllib.parse import quote
from ansible.module_utils.urls import fetch_url

socket.setdefaulttimeout(5)

# The ec2_metadata_facts module is a special case, while we generally dropped support for Python < 3.6
# this module doesn't depend on the SDK and still has valid use cases for folks working with older
# OSes.

# pylint: disable=consider-using-f-string
try:
    json_decode_error = json.JSONDecodeError
except AttributeError:
    json_decode_error = ValueError


class Ec2Metadata:
    ec2_metadata_token_uri = "http://169.254.169.254/latest/api/token"
    ec2_metadata_uri = "http://169.254.169.254/latest/meta-data/"
    ec2_metadata_instance_tags_uri = "http://169.254.169.254/latest/meta-data/tags/instance"
    ec2_sshdata_uri = "http://169.254.169.254/latest/meta-data/public-keys/0/openssh-key"
    ec2_userdata_uri = "http://169.254.169.254/latest/user-data/"
    ec2_dynamicdata_uri = "http://169.254.169.254/latest/dynamic/"

    def __init__(
        self,
        module,
        ec2_metadata_token_uri=None,
        ec2_metadata_uri=None,
        ec2_metadata_instance_tags_uri=None,
        ec2_sshdata_uri=None,
        ec2_userdata_uri=None,
        ec2_dynamicdata_uri=None,
    ):
        self.module = module
        self.uri_token = ec2_metadata_token_uri or self.ec2_metadata_token_uri
        self.uri_meta = ec2_metadata_uri or self.ec2_metadata_uri
        self.uri_instance_tags = ec2_metadata_instance_tags_uri or self.ec2_metadata_instance_tags_uri
        self.uri_user = ec2_userdata_uri or self.ec2_userdata_uri
        self.uri_ssh = ec2_sshdata_uri or self.ec2_sshdata_uri
        self.uri_dynamic = ec2_dynamicdata_uri or self.ec2_dynamicdata_uri
        self._data = {}
        self._token = None
        self._prefix = "ansible_ec2_%s"

    def _decode(self, data):
        try:
            return data.decode("utf-8")
        except UnicodeDecodeError:
            # Decoding as UTF-8 failed, return data without raising an error
            self.module.warn("Decoding user-data as UTF-8 failed, return data as is ignoring any error")
            return data.decode("utf-8", errors="ignore")

    def decode_user_data(self, data):
        is_compressed = False

        # Check if data is compressed using zlib header
        if data.startswith(b"\x78\x9c") or data.startswith(b"\x1f\x8b"):
            is_compressed = True

        if is_compressed:
            # Data is compressed, attempt decompression and decode using UTF-8
            try:
                decompressed = zlib.decompress(data, zlib.MAX_WBITS | 32)
                return self._decode(decompressed)
            except zlib.error:
                # Unable to decompress, return original data
                self.module.warn(
                    "Unable to decompress user-data using zlib, attempt to decode original user-data as UTF-8"
                )
                return self._decode(data)
        else:
            # Data is not compressed, decode using UTF-8
            return self._decode(data)

    def _fetch(self, url):
        encoded_url = quote(url, safe="%/:=&?~#+!$,;'@()*[]")
        headers = {}
        if self._token:
            headers = {"X-aws-ec2-metadata-token": self._token}

        response, info = fetch_url(self.module, encoded_url, headers=headers, force=True)

        if info.get("status") in (401, 403):
            self.module.fail_json(msg="Failed to retrieve metadata from AWS: {0}".format(info["msg"]), response=info)
        elif info.get("status") not in (200, 404):
            time.sleep(3)
            # request went bad, retry once then raise
            self.module.warn("Retrying query to metadata service. First attempt failed: {0}".format(info["msg"]))
            response, info = fetch_url(self.module, encoded_url, headers=headers, force=True)
            if info.get("status") not in (200, 404):
                # fail out now
                self.module.fail_json(
                    msg="Failed to retrieve metadata from AWS: {0}".format(info["msg"]), response=info
                )
        if response and info["status"] < 400:
            data = response.read()
            if "user-data" in encoded_url:
                return to_text(self.decode_user_data(data))
        else:
            data = None
        return to_text(data)

    def _mangle_fields(self, fields, uri, filter_patterns=None):
        filter_patterns = ["public-keys-0"] if filter_patterns is None else filter_patterns

        new_fields = {}
        for key, value in fields.items():
            split_fields = key[len(uri):].split("/")  # fmt: skip
            # Parse out the IAM role name (which is _not_ the same as the instance profile name)
            if (
                len(split_fields) == 3
                and split_fields[0:2] == ["iam", "security-credentials"]
                and ":" not in split_fields[2]
            ):
                new_fields[self._prefix % "iam-instance-profile-role"] = split_fields[2]
            if len(split_fields) > 1 and split_fields[1]:
                new_key = "-".join(split_fields)
                new_fields[self._prefix % new_key] = value
            else:
                new_key = "".join(split_fields)
                new_fields[self._prefix % new_key] = value
        for pattern in filter_patterns:
            for key in dict(new_fields):
                match = re.search(pattern, key)
                if match:
                    new_fields.pop(key)
        return new_fields

    def fetch(self, uri, recurse=True):
        raw_subfields = self._fetch(uri)
        if not raw_subfields:
            return
        subfields = raw_subfields.split("\n")
        for field in subfields:
            if field.endswith("/") and recurse:
                self.fetch(uri + field)
            if uri.endswith("/"):
                new_uri = uri + field
            else:
                new_uri = uri + "/" + field
            if new_uri not in self._data and not new_uri.endswith("/"):
                content = self._fetch(new_uri)
                if field == "security-groups" or field == "security-group-ids":
                    sg_fields = ",".join(content.split("\n"))
                    self._data["%s" % (new_uri)] = sg_fields
                else:
                    try:
                        json_dict = json.loads(content)
                        self._data["%s" % (new_uri)] = content
                        for key, value in json_dict.items():
                            self._data["%s:%s" % (new_uri, key.lower())] = value
                    except (json_decode_error, AttributeError):
                        self._data["%s" % (new_uri)] = content  # not a stringified JSON string

    def fix_invalid_varnames(self, data):
        """Change ':'' and '-' to '_' to ensure valid template variable names"""
        new_data = data.copy()
        for key, value in data.items():
            if ":" in key or "-" in key:
                newkey = re.sub(":|-", "_", key)
                new_data[newkey] = value
                del new_data[key]

        return new_data

    def fetch_session_token(self, uri_token):
        """Used to get a session token for IMDSv2"""
        metadata_token_ttl_seconds = self.module.params.get("metadata_token_ttl_seconds")
        headers = {"X-aws-ec2-metadata-token-ttl-seconds": metadata_token_ttl_seconds}
        response, info = fetch_url(self.module, uri_token, method="PUT", headers=headers, force=True)

        if info.get("status") == 403:
            self.module.fail_json(
                msg="Failed to retrieve metadata token from AWS: {0}".format(info["msg"]), response=info
            )
        elif info.get("status") not in (200, 404):
            time.sleep(3)
            # request went bad, retry once then raise
            self.module.warn("Retrying query to metadata service. First attempt failed: {0}".format(info["msg"]))
            response, info = fetch_url(self.module, uri_token, method="PUT", headers=headers, force=True)
            if info.get("status") not in (200, 404):
                # fail out now
                self.module.fail_json(
                    msg="Failed to retrieve metadata token from AWS: {0}".format(info["msg"]), response=info
                )
        if response:
            token_data = response.read()
        else:
            token_data = None
        return to_text(token_data)

    def run(self):
        self._token = self.fetch_session_token(self.uri_token)  # create session token for IMDS
        self.fetch(self.uri_meta)  # populate _data with metadata
        data = self._mangle_fields(self._data, self.uri_meta)
        data[self._prefix % "user-data"] = self._fetch(self.uri_user)
        data[self._prefix % "public-key"] = self._fetch(self.uri_ssh)

        self._data = {}  # clear out metadata in _data
        self.fetch(self.uri_dynamic)  # populate _data with dynamic data
        dyndata = self._mangle_fields(self._data, self.uri_dynamic)
        data.update(dyndata)
        data = self.fix_invalid_varnames(data)

        instance_tags_keys = self._fetch(self.uri_instance_tags)
        instance_tags_keys = instance_tags_keys.split("\n") if instance_tags_keys != "None" else []
        data[self._prefix % "instance_tags_keys"] = instance_tags_keys

        # Maintain old key for backwards compatibility
        if "ansible_ec2_instance_identity_document_region" in data:
            data["ansible_ec2_placement_region"] = data["ansible_ec2_instance_identity_document_region"]
        return data


def main():
    argument_spec = dict(
        metadata_token_ttl_seconds=dict(required=False, default=60, type="int", no_log=False),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    metadata_token_ttl_seconds = module.params.get("metadata_token_ttl_seconds")

    if metadata_token_ttl_seconds <= 0 or metadata_token_ttl_seconds > 21600:
        module.fail_json(msg="The option 'metadata_token_ttl_seconds' must be set to a value between 1 and 21600.")

    ec2_metadata_facts = Ec2Metadata(module).run()
    ec2_metadata_facts_result = dict(changed=False, ansible_facts=ec2_metadata_facts)

    module.exit_json(**ec2_metadata_facts_result)


if __name__ == "__main__":
    main()



#
# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat Inc.
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""
The iosxr l2_interfaces fact class
It is in this file the configuration is collected from the device
for a given resource, parsed, and the facts tree is populated
based on the configuration.
"""

from __future__ import absolute_import, division, print_function


__metaclass__ = type


import re

from copy import deepcopy

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common import utils

from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.argspec.l2_interfaces.l2_interfaces import (
    L2_InterfacesArgs,
)
from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.iosxr import get_os_version
from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.utils.utils import (
    Version,
    get_interface_type,
)


class L2_InterfacesFacts(object):
    """The iosxr l2_interfaces fact class"""

    def __init__(self, module, subspec="config", options="options"):
        self._module = module
        self.argument_spec = L2_InterfacesArgs.argument_spec
        spec = deepcopy(self.argument_spec)
        if subspec:
            if options:
                facts_argument_spec = spec[subspec][options]
            else:
                facts_argument_spec = spec[subspec]
        else:
            facts_argument_spec = spec

        self.generated_spec = utils.generate_dict(facts_argument_spec)

    def get_config(self, connection):
        return connection.get_config(flags="interface")

    def populate_facts(self, connection, ansible_facts, data=None):
        """Populate the facts for l2_interfaces
        :param module: the module instance
        :param connection: the device connection
        :param data: previously collected conf
        :rtype: dictionary
        :returns: facts
        """
        objs = []
        if not data:
            data = self.get_config(connection)

        # operate on a collection of resource x
        config = ("\n" + data).split("\ninterface ")
        for conf in config:
            if conf:
                obj = self.render_config(self.generated_spec, conf)
                if obj:
                    objs.append(obj)
        facts = {}
        if objs:
            facts["l2_interfaces"] = []
            params = utils.validate_config(
                self.argument_spec,
                {"config": objs},
            )
            for cfg in params["config"]:
                facts["l2_interfaces"].append(utils.remove_empties(cfg))

        ansible_facts["ansible_network_resources"].update(facts)
        return ansible_facts

    def render_config(self, spec, conf):
        """
        Render config as dictionary structure and delete keys from spec for null values
        :param spec: The facts tree, generated from the argspec
        :param conf: The configuration
        :rtype: dictionary
        :returns: The generated config
        """
        config = deepcopy(spec)
        match = re.search(r"^(\S+)", conf)
        if match:
            intf = match.group(1)

            if intf.lower() == "preconfigure":
                match = re.search(r"^(\S+) (.*)", conf)
                if match:
                    intf = match.group(2)

            if get_interface_type(intf) == "unknown":
                return {}

            config["name"] = intf

            # populate the facts from the configuration
            native_vlan = re.search(r"dot1q native vlan (\d+)", conf)
            if native_vlan:
                config["native_vlan"] = int(native_vlan.group(1))

            dot1q = utils.parse_conf_arg(conf, "encapsulation dot1q")
            os_version = get_os_version(self._module)
            if os_version and Version(os_version) > Version("7.0.0"):
                encapsulation = re.search(
                    r"encapsulation dot1q\s(\d+)\s*(second-dot1q\s\d+)?",
                    conf,
                )
                if encapsulation:
                    config["encapsulation"]["dot1q"] = int(
                        encapsulation.group(1),
                    )
                    if encapsulation.group(2):
                        config["encapsulation"]["second_dot1q"] = int(
                            encapsulation.group(2).split("second-dot1q ")[1],
                        )
            else:
                config["qvlan"] = []
                if dot1q:
                    config["qvlan"].append(dot1q.split(" ")[0])
                    if len(dot1q.split(" ")) > 1:
                        config["qvlan"].append(dot1q.split(" ")[2])

            if utils.parse_conf_cmd_arg(conf, "l2transport", True):
                config["l2transport"] = True
            if utils.parse_conf_arg(conf, "propagate"):
                config["propagate"] = True
            config["l2protocol"] = []

            cdp = utils.parse_conf_arg(conf, "l2protocol cdp")
            pvst = utils.parse_conf_arg(conf, "l2protocol pvst")
            stp = utils.parse_conf_arg(conf, "l2protocol stp")
            vtp = utils.parse_conf_arg(conf, "l2protocol vtp")
            cpsv = utils.parse_conf_arg(conf, "l2protocol cpsv")
            if cdp:
                config["l2protocol"].append({"cdp": cdp})
            if pvst:
                config["l2protocol"].append({"pvst": pvst})
            if stp:
                config["l2protocol"].append({"stp": stp})
            if vtp:
                config["l2protocol"].append({"vtp": vtp})
            if cpsv:
                config["l2protocol"].append({"cpsv": cpsv})

            return utils.remove_empties(config)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: device_administration_local_exception_rules_reset_hitcount
short_description: Resource module for Device Administration Local Exception Rules Reset Hitcount
description:
- Manage operation create of the resource Device Administration Local Exception Rules Reset Hitcount.
- Device Admin - Reset HitCount for local exceptions.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.ise.module
author: Rafael Campos (@racampos)
options:
  policyId:
    description: PolicyId path parameter. Policy id.
    type: str
requirements:
- ciscoisesdk >= 2.2.3
- python >= 3.5
seealso:
- name: Cisco ISE documentation for Device Administration - Authorization Exception Rules
  description: Complete reference of the Device Administration - Authorization Exception Rules API.
  link: https://developer.cisco.com/docs/identity-services-engine/v1/#!policy-openapi
notes:
  - SDK Method used are
    device_administration_authorization_exception_rules.DeviceAdministrationAuthorizationExceptionRules.reset_hit_counts_device_admin_local_exceptions,

  - Paths used are
    post /device-admin/policy-set/{policyId}/exception/reset-hitcount,

"""

EXAMPLES = r"""
- name: Create
  cisco.ise.device_administration_local_exception_rules_reset_hitcount:
    ise_hostname: "{{ise_hostname}}"
    ise_username: "{{ise_username}}"
    ise_password: "{{ise_password}}"
    ise_verify: "{{ise_verify}}"
    policyId: string

"""

RETURN = r"""
ise_response:
  description: A dictionary or list with the response returned by the Cisco ISE Python SDK
  returned: always
  type: dict
  sample: >
    {
      "message": "string"
    }
"""



#!/usr/bin/python

# Copyright: (c) 2012, Stephen Fromm <sfromm@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = r'''
---
module: seboolean
short_description: Toggles SELinux booleans
description:
     - Toggles SELinux booleans.
version_added: "1.0.0"
options:
  name:
    description:
      - Name of the boolean to configure.
    required: true
    type: str
  persistent:
    description:
      - Set to V(true) if the boolean setting should survive a reboot.
    type: bool
    default: false
  state:
    description:
      - Desired boolean value.
    type: bool
    required: true
  ignore_selinux_state:
    description:
    - Useful for scenarios (chrooted environment) that you can't get the real SELinux state.
    type: bool
    default: false
notes:
   - Not tested on any Debian based system.
requirements:
- libselinux-python
- libsemanage-python
- python3-libsemanage
author:
- Stephen Fromm (@sfromm)
'''

EXAMPLES = r'''
- name: Set httpd_can_network_connect flag on and keep it persistent across reboots
  ansible.posix.seboolean:
    name: httpd_can_network_connect
    state: true
    persistent: true
'''

import os
import traceback

SELINUX_IMP_ERR = None
try:
    import selinux
    HAVE_SELINUX = True
except ImportError:
    SELINUX_IMP_ERR = traceback.format_exc()
    HAVE_SELINUX = False

SEMANAGE_IMP_ERR = None
try:
    import semanage
    HAVE_SEMANAGE = True
except ImportError:
    SEMANAGE_IMP_ERR = traceback.format_exc()
    HAVE_SEMANAGE = False

from ansible.module_utils.basic import AnsibleModule, missing_required_lib
from ansible.module_utils._text import to_text
from ansible_collections.ansible.posix.plugins.module_utils._respawn import respawn_module, HAS_RESPAWN_UTIL


def get_runtime_status(ignore_selinux_state=False):
    return True if ignore_selinux_state is True else selinux.is_selinux_enabled()


def get_boolean_value(module, name):
    state = 0
    try:
        state = selinux.security_get_boolean_active(name)
    except OSError:
        module.fail_json(msg="Failed to determine current state for boolean %s" % name)
    if state == 1:
        return True
    else:
        return False


def semanage_get_handle(module):
    handle = semanage.semanage_handle_create()
    if not handle:
        module.fail_json(msg="Failed to create semanage library handle")

    managed = semanage.semanage_is_managed(handle)
    if managed <= 0:
        semanage.semanage_handle_destroy(handle)
    if managed < 0:
        module.fail_json(msg="Failed to determine whether policy is manage")
    if managed == 0:
        if os.getuid() == 0:
            module.fail_json(msg="Cannot set persistent booleans without managed policy")
        else:
            module.fail_json(msg="Cannot set persistent booleans; please try as root")

    if semanage.semanage_connect(handle) < 0:
        semanage.semanage_handle_destroy(handle)
        module.fail_json(msg="Failed to connect to semanage")

    return handle


def semanage_begin_transaction(module, handle):
    if semanage.semanage_begin_transaction(handle) < 0:
        semanage.semanage_handle_destroy(handle)
        module.fail_json(msg="Failed to begin semanage transaction")


def semanage_set_boolean_value(module, handle, name, value):
    rc, t_b = semanage.semanage_bool_create(handle)
    if rc < 0:
        semanage.semanage_handle_destroy(handle)
        module.fail_json(msg="Failed to create seboolean with semanage")

    if semanage.semanage_bool_set_name(handle, t_b, name) < 0:
        semanage.semanage_handle_destroy(handle)
        module.fail_json(msg="Failed to set seboolean name with semanage")

    rc, boolkey = semanage.semanage_bool_key_extract(handle, t_b)
    if rc < 0:
        semanage.semanage_handle_destroy(handle)
        module.fail_json(msg="Failed to extract boolean key with semanage")

    rc, exists = semanage.semanage_bool_exists(handle, boolkey)
    if rc < 0:
        semanage.semanage_handle_destroy(handle)
        module.fail_json(msg="Failed to check if boolean is defined")
    if not exists:
        semanage.semanage_handle_destroy(handle)
        module.fail_json(msg="SELinux boolean %s is not defined in persistent policy" % name)

    rc, sebool = semanage.semanage_bool_query(handle, boolkey)
    if rc < 0:
        semanage.semanage_handle_destroy(handle)
        module.fail_json(msg="Failed to query boolean in persistent policy")

    semanage.semanage_bool_set_value(sebool, value)

    if semanage.semanage_bool_modify_local(handle, boolkey, sebool) < 0:
        semanage.semanage_handle_destroy(handle)
        module.fail_json(msg="Failed to modify boolean key with semanage")

    if (
            selinux.is_selinux_enabled()
            and semanage.semanage_bool_set_active(handle, boolkey, sebool) < 0
    ):
        semanage.semanage_handle_destroy(handle)
        module.fail_json(msg="Failed to set boolean key active with semanage")

    semanage.semanage_bool_key_free(boolkey)
    semanage.semanage_bool_free(t_b)
    semanage.semanage_bool_free(sebool)


def semanage_get_boolean_value(module, handle, name):
    rc, t_b = semanage.semanage_bool_create(handle)
    if rc < 0:
        semanage.semanage_handle_destroy(handle)
        module.fail_json(msg="Failed to create seboolean with semanage")

    if semanage.semanage_bool_set_name(handle, t_b, name) < 0:
        semanage.semanage_handle_destroy(handle)
        module.fail_json(msg="Failed to set seboolean name with semanage")

    rc, boolkey = semanage.semanage_bool_key_extract(handle, t_b)
    if rc < 0:
        semanage.semanage_handle_destroy(handle)
        module.fail_json(msg="Failed to extract boolean key with semanage")

    rc, exists = semanage.semanage_bool_exists(handle, boolkey)
    if rc < 0:
        semanage.semanage_handle_destroy(handle)
        module.fail_json(msg="Failed to check if boolean is defined")
    if not exists:
        semanage.semanage_handle_destroy(handle)
        module.fail_json(msg="SELinux boolean %s is not defined in persistent policy" % name)

    rc, sebool = semanage.semanage_bool_query(handle, boolkey)
    if rc < 0:
        semanage.semanage_handle_destroy(handle)
        module.fail_json(msg="Failed to query boolean in persistent policy")

    value = semanage.semanage_bool_get_value(sebool)

    semanage.semanage_bool_key_free(boolkey)
    semanage.semanage_bool_free(t_b)
    semanage.semanage_bool_free(sebool)

    return value


def semanage_commit(module, handle, load=0):
    semanage.semanage_set_reload(handle, load)
    if semanage.semanage_commit(handle) < 0:
        semanage.semanage_handle_destroy(handle)
        module.fail_json(msg="Failed to commit changes to semanage")


def semanage_destroy_handle(module, handle):
    rc = semanage.semanage_disconnect(handle)
    semanage.semanage_handle_destroy(handle)
    if rc < 0:
        module.fail_json(msg="Failed to disconnect from semanage")


# The following method implements what setsebool.c does to change
# a boolean and make it persist after reboot..
def semanage_boolean_value(module, name, state):
    value = 0
    changed = False
    if state:
        value = 1
    try:
        handle = semanage_get_handle(module)
        semanage_begin_transaction(module, handle)
        cur_value = semanage_get_boolean_value(module, handle, name)
        if cur_value != value:
            changed = True
            if not module.check_mode:
                semanage_set_boolean_value(module, handle, name, value)
                semanage_commit(module, handle)
        semanage_destroy_handle(module, handle)
    except Exception as e:
        module.fail_json(msg=u"Failed to manage policy for boolean %s: %s" % (name, to_text(e)))
    return changed


def set_boolean_value(module, name, state):
    rc = 0
    value = 0
    if state:
        value = 1
    try:
        rc = selinux.security_set_boolean(name, value)
    except OSError:
        module.fail_json(msg="Failed to set boolean %s to %s" % (name, value))
    if rc == 0:
        return True
    else:
        return False


def main():
    module = AnsibleModule(
        argument_spec=dict(
            ignore_selinux_state=dict(type='bool', default=False),
            name=dict(type='str', required=True),
            persistent=dict(type='bool', default=False),
            state=dict(type='bool', required=True),
        ),
        supports_check_mode=True,
    )

    if not HAVE_SELINUX and not HAVE_SEMANAGE and HAS_RESPAWN_UTIL:
        # Only respawn the module if both libraries are missing.
        # If only one is available, then usage of the "wrong" (i.e. not the system one)
        # python interpreter is likely not the problem.
        respawn_module("selinux")

    if not HAVE_SELINUX:
        module.fail_json(msg=missing_required_lib('libselinux-python'), exception=SELINUX_IMP_ERR)

    if not HAVE_SEMANAGE:
        module.fail_json(msg=missing_required_lib('libsemanage-python or python3-libsemanage'), exception=SEMANAGE_IMP_ERR)

    ignore_selinux_state = module.params['ignore_selinux_state']

    if not get_runtime_status(ignore_selinux_state):
        module.fail_json(msg="SELinux is disabled on this host.")

    name = module.params['name']
    persistent = module.params['persistent']
    state = module.params['state']

    result = dict(
        name=name,
        persistent=persistent,
        state=state
    )
    changed = False

    if hasattr(selinux, 'selinux_boolean_sub'):
        # selinux_boolean_sub allows sites to rename a boolean and alias the old name
        # Feature only available in selinux library since 2012.
        name = selinux.selinux_boolean_sub(name)

    if persistent:
        changed = semanage_boolean_value(module, name, state)
    elif selinux.is_selinux_enabled():
        cur_value = get_boolean_value(module, name)
        if cur_value != state:
            changed = True
            if not module.check_mode:
                changed = set_boolean_value(module, name, state)
                if not changed:
                    module.fail_json(msg="Failed to set boolean %s to %s" % (name, state))
                try:
                    selinux.security_commit_booleans()
                except Exception:
                    module.fail_json(msg="Failed to commit pending boolean %s value" % name)

    result['changed'] = changed

    module.exit_json(**result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: wireless_controllers_ssid_details_count_v1_info
short_description: Information module for Wireless Controllers Ssid Details Count V1
description:
- Get all Wireless Controllers Ssid Details Count V1.
- Retrieves the count of SSIDs associated with the specific Wireless Controller.
version_added: '6.15.0'
extends_documentation_fragment:
  - cisco.dnac.module_info
author: Rafael Campos (@racampos)
options:
  headers:
    description: Additional headers.
    type: dict
  networkDeviceId:
    description:
    - >
      NetworkDeviceId path parameter. Obtain the network device ID value by using the API call GET
      /dna/intent/api/v1/network-device/ip-address/${ipAddress}.
    type: str
  adminStatus:
    description:
    - >
      AdminStatus query parameter. Utilize this query parameter to obtain the number of SSIDs according to their
      administrative status. A 'true' value signifies that the admin status of the SSID is enabled, while a
      'false' value indicates that the admin status of the SSID is disabled.
    type: bool
  managed:
    description:
    - >
      Managed query parameter. If value is 'true' means SSIDs are configured through design.If the value is
      'false' means out of band configuration from the Wireless Controller.
    type: bool
requirements:
- dnacentersdk >= 2.4.9
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Wireless GetSSIDCountForSpecificWirelessControllerV1
  description: Complete reference of the GetSSIDCountForSpecificWirelessControllerV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!get-ssid-count-for-specific-wireless-controller
notes:
  - SDK Method used are
    wireless.Wireless.get_ssid_count_for_specific_wireless_controller_v1,

  - Paths used are
    get /dna/intent/api/v1/wirelessControllers/{networkDeviceId}/ssidDetails/count,

"""

EXAMPLES = r"""
- name: Get all Wireless Controllers Ssid Details Count V1
  cisco.dnac.wireless_controllers_ssid_details_count_v1_info:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    headers: "{{my_headers | from_json}}"
    adminStatus: True
    managed: True
    networkDeviceId: string
  register: result

"""
RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": {
        "count": 0
      },
      "version": "string"
    }
"""
